# -*- coding: utf-8 -*-

from __future__ import print_function
import os
import re
import sys
import json
import datetime
import shutil
import subprocess
import time
import signal
import zipfile
from multiprocessing import Process, Queue
from functools import wraps
from threading import Timer
import platform
from . import base
from .base import log, pjoin

def lemon(conf = None):
	base.check_install('pyside')
	if not conf:
		if base.conf.folder == 'problem':
			raise Exception('Can\'t dump a single problem to lemon, try to dump a day or a contest.')
		base.remkdir('lemon')
		if base.conf.folder == 'day':
			lemon(base.conf)
		else:
			for day in base.days():
				os.makedirs(base.pjoin('lemon', day.route))
				lemon(day)
		return
	log.info(u'导出lemon工程：%s' % conf.route)
	os.makedirs(base.pjoin('lemon', conf.route, 'data'))
	os.makedirs(base.pjoin('lemon', conf.route, 'source'))
	jzp_magic = 0x20111127
	zlib_magic = 0x185E
	import zlib
	from PySide import QtCore
	obuff = QtCore.QByteArray()
	ost = QtCore.QDataStream(obuff, QtCore.QIODevice.WriteOnly)
	ost.writeQString(conf['name'])
	probs = list(conf.probs())
	ost.writeInt32(len(probs))
	for prob in probs:
		log.info(u'导出lemon题目：%s' % prob.route)
		ost.writeQString(base.tr(prob['title']))
		ost.writeQString(prob['name'])
		ost.writeQString(prob['name'] + '.in')
		ost.writeQString(prob['name'] + '.out')
		ost.writeBool(False)
		ost.writeBool(False)
		ost.writeInt32(1 if prob['type'] == 'output' else 0)
		ost.writeInt32(1)		# Judge Type TODO: What if there is spj (code = 4)
		ost.writeQString('--ignore-space-change --text --brief')
		ost.writeInt32(3)		# real precision (float number error bound)
		ost.writeQString('')	# spj route TODO: What if there is spj
		ost.writeInt32(len([i for i in prob.get('compile', {}) if i in {'cpp', 'c', 'pas'}]))
		for key, val in prob.get('compile', {}).items():
			try:
				ost.writeQString({
					'cpp' : 'g++',
					'c' : 'gcc',
					'pas' : 'fpc'
				}[key])
				ost.writeQString(val)
			except:
				pass
		ost.writeQString('out')
		if prob.packed:
			ost.writeInt32(len(prob['data']))
			for datum in prob['data']:
				ost.writeInt32(datum['score'])
				ost.writeInt32(datum.get('time limit', prob.get('time limit', 0)) * 1000)
				ost.writeInt32(base.Memory(datum.get('memory limit', prob.get('memory limit', '0 B'))).MB)
				tc = datum['cases']
				ost.writeInt32(len(tc))
				for c in tc:
					ost.writeQString(base.pjoin(prob['name'], str(c) + '.in'))
				ost.writeInt32(len(tc))
				for c in tc:
					ost.writeQString(base.pjoin(prob['name'], str(c) + '.ans'))
		else:
			score = (100. / len(prob.test_cases) if len(prob.test_cases) > 0 else 0.)
			ost.writeInt32(len(prob.test_cases))
			for c in prob.test_cases:
				ost.writeInt32(score)
				ost.writeInt32(prob['time limit'] * 1000)
				ost.writeInt32(prob.memory_limit().MB)
				ost.writeInt32(1)
				ost.writeQString(base.pjoin(prob['name'], str(c) + '.in'))
				ost.writeInt32(1)
				ost.writeQString(base.pjoin(prob['name'], str(c) + '.ans'))
		
		sp = list(prob.route.split('/'))
		target = ['lemon'] + sp[:-1] + ['data', sp[-1]]
		shutil.copytree(base.pjoin(prob.path, 'data'), base.pjoin(*target))
	
	compressed = QtCore.QByteArray(zlib.compress(str(obuff)))
	obuff = QtCore.QByteArray()
	ost = QtCore.QDataStream(obuff, QtCore.QIODevice.WriteOnly)
	ost.writeUInt32(zlib_magic)
	ost.writeRawData(str(compressed))
	file_ = QtCore.QFile(base.pjoin('lemon', conf.route, conf['name'] + '.cdf'))
	file_.open(QtCore.QIODevice.WriteOnly)
	ofs = QtCore.QDataStream(file_)
	ofs.writeUInt32(jzp_magic)
	ofs.writeUInt16(QtCore.qChecksum(str(obuff), len(obuff)))
	ofs.writeUInt32(len(obuff))
	ofs.writeRawData(str(obuff))
	file_.close()
	
	base.run_r(base.unix2dos, base.pjoin('lemon', conf.route, 'data'))
	
	if base.do_zip:
		import zipfile
		with zipfile.ZipFile(base.pjoin('lemon', conf.route) + '.zip', 'w') as z:
			base.run_r(lambda path : z.write(path), base.pjoin('lemon', conf.route))
	log.warning(u'目前SPJ的支持暂时还没有实现，有需要请手工配置。')
	log.warning(u'目前lemon的编译选项是写在注册表里的，暂时没有实现该功能，请手工配置。')

def arbiter_main(conf = None,daynum = 0):
	def arbiter_info(info, filename):
		with open(filename,'wb') as ofile:
			for key, val in info.items():
				ofile.write(('%s%s\n' % (key, val)).encode('gbk'))
	if not conf:
		log.info('makedirs')
		if not os.path.exists('arbiter'):
			base.remkdir('arbiter')
		base.remkdir(pjoin('arbiter', 'main'))
		os.makedirs(base.pjoin('arbiter', 'main','data'))
		os.makedirs(base.pjoin('arbiter', 'main','final'))
		os.makedirs(base.pjoin('arbiter', 'main','players'))
		os.makedirs(base.pjoin('arbiter', 'main','result'))
		os.makedirs(base.pjoin('arbiter', 'main','filter'))
		os.makedirs(base.pjoin('arbiter', 'main','tmp'))
		if base.conf.folder == 'problem':
			raise Exception('Can\'t dump a single problem to arbiter, try to dump a day or a contest.')
		if base.conf.folder == 'day':
			arbiter(base.days())
		else:
			for idx, day in enumerate(base.days(), start = 1):
				os.makedirs(base.pjoin('arbiter', 'main','players',day.route))
				os.makedirs(base.pjoin('arbiter', 'main','result',day.route))
				arbiter_main(day,idx)
			log.info('dos2unix')
			base.run_r(base.dos2unix, base.pjoin('arbiter', 'main', 'data'))
			shutil.copytree(base.pjoin('arbiter', 'main','data'),base.pjoin('arbiter', 'main','evaldata'))	#这里也不能直接copy，见下面的处理方式
			cfg = {}
			cfg['NAME='] = base.conf['name']
			cfg['DAYNUM='] = idx
			cfg['ENV='] = 'env.info'
			cfg['PLAYER='] = 'player.info'
			cfg['TEAM='] = 'team.info'
			cfg['MISC='] = 'misc.info'
			arbiter_info(cfg,base.pjoin('arbiter', 'main','setup.cfg'))
			team = {}
			arbiter_info(team,base.pjoin('arbiter', 'main','team.info'))
			'''arbiter_info(userlist,base.pjoin('arbiter','player.info'))'''
		return
	dayinfo = {}
	dayinfo['NAME='] = u'第'+str(daynum)+u'场'+u'--机试'
	dayinfo['PLAYERDIR='] = ''
	dayinfo['CASEDIR='] = ''
	dayinfo['BASESCORE='] = 0
	dayinfo['TASKNUM='] = len(conf['subdir'])
	arbiter_info(dayinfo,base.pjoin('arbiter', 'main', conf['name']+'.info'))
	for probnum, prob in enumerate(conf.sub, start = 1):
		log.info(prob['name'])
		probinfo = {}
		probinfo['TITLE='] = ''
		probinfo['NAME='] = prob['name']
		probinfo['RUN='] = ''
		probinfo['INFILESUFFIX='] = 'in'
		probinfo['ANSFILESUFFIX='] = 'ans'
		probinfo['PLUG='] = prob['name']+'_e'
		if prob['type'] == 'program':
			probinfo['TYPE='] = 'SOURCE'
		else:
			log.warning(u'暂时只支持非交互式程序题。')
		probinfo['LIMIT='] = int(prob['time limit'])
		probinfo['MEMLIMITS='] = int(prob.memory_limit().MB)
		probinfo['SAMPLES='] = len(prob.test_cases)
		score_per_case = 100 // len(prob.test_cases)
		if score_per_case * len(prob.test_cases) != 100:
			log.info(u'满分不是100哦。')
		probinfo['CCL=c@gcc'] = ' -o %o %i ' + prob['compile']['c']
		probinfo['CCL=cpp@g++'] = ' -o %o %i ' + prob['compile']['cpp']
		probinfo['CCL=pas@fpc'] = ' %i ' + prob['compile']['pas']
		if prob.packed:
			raise Exception('Can\'t dump packed problem for arbiter.')
		for idx, case in enumerate(prob.test_cases, start = 1):
			'''print('copyfile %s'%base.pjoin(prob.path,'data',case+'.in'))'''
			shutil.copy(
				base.pjoin(prob.path,'data',str(case)+'.in'),
				base.pjoin('arbiter', 'main','data',prob['name']+str(idx)+'.in')
			)
			'''print('copyfile %s'%base.pjoin(prob.path,'data',case+'.ans'))'''
			shutil.copy(
				base.pjoin(prob.path,'data',str(case)+'.ans'),
				base.pjoin('arbiter', 'main','data',prob['name']+str(idx)+'.ans')
			)
			probinfo['MARK='+str(idx)+'@'] = str(int(score_per_case))
		'''for idx, userdir in enumerate(prob['users'],start = 1):
			for idx2, code in enumerate(prob['users'][userdir],start = 1):
				dirname = prob['name']+'-'+str(idx)+str(idx2)
				codename = userdir + code
				tmplist = prob['users'][userdir][code].split('/')
				codedir = base.pjoin(prob.path,*tmplist)
				os.makedirs(base.pjoin('arbiter','players',conf['name'],dirname,prob['name']))
				shutil.copy(codedir,base.pjoin('arbiter','players',conf['name'],dirname,prob['name']))
				userlist[dirname + '@'] =  codename'''
		shutil.copy(base.pjoin(base.path,'sample','arbiter_e.sample'),base.pjoin('arbiter', 'main','filter',prob['name']+'_e'))
		arbiter_info(probinfo,base.pjoin('arbiter', 'main','task'+str(daynum)+'_'+str(probnum)+'.info'))

def arbiter_down(conf = None):
	if not conf:
		conf = base.conf
		if not os.path.exists('arbiter'):
			base.remkdir('arbiter')
		base.remkdir(pjoin('arbiter', 'down'))
	if conf.folder == 'problem':
		raise Exception('Can\'t dump a single problem to arbiter, try to dump a day or a contest.')
	if conf.folder == 'contest':
		for idx, day in enumerate(base.days(), start = 1):
			os.makedirs(base.pjoin(pjoin('arbiter', 'down'), day['name']))
			arbiter_down(day)
		log.info('dos2unix')
		base.run_r(base.dos2unix, base.pjoin(pjoin('arbiter', 'down')))
		return
	for prob in conf.probs():
		log.info(prob.route)
		os.makedirs(base.pjoin(pjoin('arbiter', 'down'), prob.route))
		for idx, case in enumerate(prob.sample_cases, start = 1):
			shutil.copy(
				base.pjoin(prob.path,'down',str(case)+'.in'),
				base.pjoin('arbiter', 'down', prob.route, prob['name']+str(idx)+'.in')
			)
			shutil.copy(
				base.pjoin(prob.path,'down',str(case)+'.ans'),
				base.pjoin('arbiter', 'down', prob.route, prob['name']+str(idx)+'.ans')
			)

def arbiter():
	base.remkdir('arbiter')
	arbiter_main()
	arbiter_down()

def tsinsen_oj():
	import random
	if type(base.conf) != base.Problem:
		log.error(u'只能转换一道题目，请到相应题目目录下运行')
		return
	hash_ch = list(map(
		chr,
		[ord('a') + i for i in range(26)] + \
		[ord('A') + i for i in range(26)] + \
		[ord('0') + i for i in range(10)]
	))
	hash = lambda l = 10 : ''.join([hash_ch[random.randint(0, len(hash_ch) - 1)] for i in range(l)])
	if not os.path.exists('tsinsen-oj'):
		os.makedirs('tsinsen-oj')
	for day in base.days():
		p = base.pjoin('tsinsen-oj', day.route)
		if not os.path.exists(p):
			os.makedirs(p)
	for prob in base.probs():
		p = base.pjoin('tsinsen-oj', prob.route)
		if not os.path.exists(p):
			os.makedirs(p)
		if os.path.exists(base.pjoin(prob.path, 'down')):
			with zipfile.ZipFile(base.pjoin('tsinsen-oj', prob.route, 'down.zip'), 'w') as z:
				base.run_r(lambda path : z.write(path), base.pjoin(prob.path, 'down'))
			prob.tsinsen_down = hash(8)
		else:
			prob.tsinsen_down = None
		files = {}
		path = base.pjoin(prob.path, 'resources')
		if os.path.exists(path):
			base.run_r(lambda p : files.__setitem__(p[len(path) + 1:], hash(8)), path)
		prob.tsinsen_files = files
	if base.do_render:
		from . import ren
		tmp = base.start_file
		base.start_file = False
		ren.Html('tsinsen-oj').run()
		base.start_file = tmp
	else:
		log.warning(u'如果你使用了文件，不重新渲染题面会导致tsinsen的文件失效。')
	
	read_file = lambda path : open(path, 'rb').read().decode('utf-8')
	
	Title = lambda : prob.tr('title')
	CheckPoint = lambda : prob['key words'] if 'key words' in prob else ''
	TestMethod = lambda : 'DEFAULT'
	InputFileName = lambda : ''
	OutputFileName = lambda : ''
	TimeLimit = lambda : '%.1fs' % prob['time limit']
	MemoryLimit = lambda : '%.1fMB' % prob.memory_limit().MB
	
	def Checkers():
		path = base.pjoin('data', 'chk', 'chk.cpp')
		if os.path.exists(path):
			return open(path, 'rb').read().decode('utf-8')
		else:
			return '\n'
	def Description():
		path = pjoin('statements', 'tsinsen-oj', prob.route if prob.route != '' else prob['name']) + '.html'
		if not os.path.exists(path):
			log.warning(u'找不到题面，你可能需要自己手工添加题面。')
			return ''
		header = u'<p>下载目录：<img src="/RequireFile.do?fid=%s" alt="另存为图片下载" />（另存为图片下载）</p>\n' % prob.tsinsen_down
		return header + read_file(path)
	def Solution():
		for user, codes in prob['users'].items():
			for name, path in codes.items():
				if prob.expect(user, name, 100):
					return read_file(path['path'])
		for user, codes in prob['users'].items():
			for name, path in codes.items():
				return read_file(path)
		return ''
	def add_shell(name, func):
		h = hash()
		ret = '%s=\n' % name
		ret += '======================%s\n' % h
		res = func()
		if not res.endswith('\n'):
			res += '\n'
		ret += res
		ret += '======================%s\n' % h
		return ret.encode('utf-8')
	def to_base64(path):
		import base64
		s = base64.b64encode(open(path, 'rb').read()).decode('utf-8')
		return '\n'.join(s[pos:pos+76] for pos in range(0, len(s), 76))
	tokens = [
		'Title', 'CheckPoint', 'Checkers', 'TestMethod', 'Description',
		'InputFileName', 'OutputFileName', 'Solution', 'TimeLimit', 'MemoryLimit'
	]
	for prob in base.probs():
		result_file = base.pjoin('tsinsen-oj', prob.route) + '.txt'
		with open(result_file, 'wb') as f:
			for token in tokens:
				f.write(add_shell(token, eval(token)))
			if prob.packed:
				log.warning(u'清橙OJ不支持打包评测和指定测试点分值，直接将所有测试点视为相同。')
			for datum in prob.test_cases:
				f.write(add_shell('InData', lambda : read_file(base.pjoin('data', datum + '.in'))))
				f.write(add_shell('OutData', lambda : read_file(base.pjoin('data', datum + '.ans'))))
			if prob.tsinsen_down:
				f.write(add_shell('FileName', lambda : 'down.zip'))
				f.write(add_shell('File(%s)' % prob.tsinsen_down, lambda : to_base64(base.pjoin('tsinsen-oj', prob.route, 'down.zip'))))
			for key, val in prob.tsinsen_files.items():
				f.write(add_shell('FileName', lambda : key.replace('/', '-').replace('\\', '-')))
				f.write(add_shell('File(%s)' % val, lambda : to_base64(base.pjoin(prob.path, 'resources', key))))
		if base.start_file:
			base.xopen_file(result_file)

def tuoj_down(conf = None):
	if not conf:
		conf = base.conf
		if not os.path.exists('tuoj'):
			base.remkdir('tuoj')
		base.remkdir(pjoin('tuoj', 'down'))
	if conf.folder == 'contest':
		for day in base.days():
			os.makedirs(base.pjoin(pjoin('tuoj', 'down'), day['name']))
			tuoj_down(day)
		return
	for prob in conf.probs():
		log.info(prob.route)
		os.makedirs(base.pjoin(pjoin('tuoj', 'down'), prob.route))
		for idx, case in enumerate(prob.sample_cases, start = 1):
			shutil.copy(
				base.pjoin(prob.path, 'down', str(case) + '.in'),
				base.pjoin('tuoj', 'down', prob.route, str(case) + '.in')
			)
			shutil.copy(
				base.pjoin(prob.path, 'down',str(case) + '.ans'),
				base.pjoin('tuoj', 'down', prob.route, str(case) + '.ans')
			)
		base.run_r(base.dos2unix, base.pjoin(pjoin('tuoj', 'down', prob.route)))

work_list = {
	'lemon' : lemon,
	'arbiter' : arbiter,
	'arbiter-main' : arbiter_main,
	'arbiter-down' : arbiter_down,
	'tsinsen-oj' : tsinsen_oj,
	'tuoj-down' : tuoj_down
}

if __name__ == '__main__':
	try:
		if base.init():
			for base.work in base.works:
				base.run_exc(work_list[base.work])
		else:
			log.info(u'这个工具用于导出成其他类型的工程。')
			log.info(u'支持的工作：%s' % ','.join(sorted(work_list.keys())))
	except base.NoFileException as e:
		log.error(e)
		log.info(u'尝试使用`python -m tuack.gen -h`获取如何生成一个工程。')



# -*- coding: utf-8 -*-

from __future__ import print_function
import os
import re
import sys
import json
import datetime
import shutil
import subprocess
import time
import signal
import zipfile
from multiprocessing import Process, Queue
from functools import wraps
from threading import Timer
import platform
from . import base
from .base import log, pjoin, rjoin
import traceback

def time2float(inp):
	m = re.match(r'(\d*)(:|h)(\d*)(:|m)(\d*\.?\d*)s?', inp)
	if m:
		return float(m.group(1)) * 3600 + float(m.group(3)) * 60 + float(m.group(5))
	m = re.match(r'(\d*)(:|m)(\d*\.?\d*)s?', inp)
	if m:
		return float(m.group(1)) * 60 + float(m.group(3))
	m = re.match(r'(\d*\.?\d*)s?', inp)
	if m:
		return float(m.group(1))
	raise Exception('Timer broken')

def run_windows(name, tl, ml, input = None, output = None, vm = None):
	'''
	On windows, memory limit is not considered.
	'''
	try:
		fin = (open(input) if input else None)
		fout = (open(output, 'w') if output else None)
		t = time.clock()
		pro = subprocess.Popen(vm(name, ml) if vm else name, stdin = fin, stdout = fout)
		if fout:
			fout.close()
		if fin:
			fin.close()
	except:
		return 'Can\'t run program.', 0.0
	while True:
		ret = pro.poll()
		if ret != None:
			t = time.clock() - t
			if ret == 0:
				ret = None
			else:
				ret = 'Runtime error %d.' % ret
			break
		if (time.clock() - t) >= tl * base.time_multiplier:
			pro.kill()
			t = 0.0
			ret = 'Time out.'
			break
		time.sleep(1e-2)
	time.sleep(1e-2)
	return ret, t

def runner_linux(name, que, ml, input = None, output = None, vm = None):
	pro = subprocess.Popen(
		'%s time -f "%%%s" -o timer %s %s %s' % (
			'ulimit -v %d;' % int(base.Memory(ml).KB) if vm != base.runners['java'] else '',
			'U' if base.user_time else 'E',
			vm(name, ml) if vm else './%s' % name,
			'< %s' % input if input else '',
			'> %s' % output if output else '',
		),
		shell = True,
		preexec_fn = os.setsid
	)
	que.put(pro.pid)
	ret = pro.wait()
	que.put(ret)
	
def run_linux(name, tl, ml, input = None, output = None, vm = None):
	que = Queue()
	pro = Process(target = runner_linux, args = (name, que, ml, input, output, vm))
	pro.start()
	pro.join(tl * base.time_multiplier)
	if que.qsize() == 0:
		log.error('Runner broken.')
	elif que.qsize() == 1:
		ret = 'Time out.'
		pid = que.get()
		os.killpg(os.getpgid(pid), signal.SIGTERM)
		t = 0.
	else:
		pid = que.get()
		ret = que.get()
		if ret == 0:
			try:
				t = time2float(open('timer').readline())
			except Exception as e:
				log.debug(e)
				log.warning('Timer broken.')
				t = 0.
			ret = None
		else:
			ret = 'Runtime error %d.(MLE will cause RE as well in linux)' % ret
			t = 0.
	return ret, t

def runner_mac(name, que, ml, input = None, output = None, vm = None):
	pro = subprocess.Popen(
		'ulimit -v %d; (time -p %s %s %s) 2> timer' % (
			int(base.Memory(ml).KB),
			vm(name, ml) if vm else './%s' % name,
			'< %s' % input if input else '',
			'> %s' % output if output else '',
		),
		shell = True,
		preexec_fn = os.setsid
	)
	que.put(pro.pid)
	ret = pro.wait()
	que.put(ret)

def run_mac(name, tl, ml, input = None, output = None, vm = None):
	que = Queue()
	pro = Process(target = runner_mac, args = (name, que, ml, input, output, vm))
	pro.start()
	pro.join(tl * base.time_multiplier)
	if que.empty():
		log.error('Runner broken.')
	else:
		pid = que.get()
		if que.empty():
			ret = 'Time out.'
			os.killpg(os.getpgid(pid), signal.SIGTERM)
			t = 0.
		else:
			ret = que.get()
			if ret == 0:
				try:
					with open('timer') as f:
						for i in range(2 if base.user_time else 1):
							line = f.readline()
						t = time2float(line.strip().split()[-1])
				except:
					base.warning('Timer broken.')
					t = 0.
				ret = None
			else:
				ret = 'Runtime error %d.' % ret
				t = 0.
	return ret, t

if base.system == 'Linux':
	run = run_linux
elif base.system == 'Windows':
	run = run_windows
elif base.system == 'Darwin':
	run = run_mac
else:
	run = run_linux
	log.warning(u'未知的操作系统，尝试当做Linux运行。')
	
def compile(prob, name):
	for lang, args in prob['compile'].items():
		if os.path.exists(pjoin('tmp', name + '.' + lang)):
			base.check_install(lang)
			os.chdir('tmp')
			try:
				ret = subprocess.call(
					base.compilers[lang](name, args, base.macros[base.work], ml = prob['memory limit']),
					shell = True,
					stdout = open('stdout', 'w'),
					stderr = open('stderr', 'w')
				)
			except Exception as e:
				with open('stderr', 'a') as f:
					f.write(str(e))
				ret = -1
			os.chdir('..')
			if ret:
				log.info('`' + name + '.' + lang + u'`编译失败，详情见`compile.log`。')
				with open('compile.log', 'a') as f:
					import __main__
					f.write(u'## 脚本%s/，工程路径%s，参数%s，开始于%s。\n' % (__main__.__file__, os.getcwd(), str(sys.argv[1:]), str(datetime.datetime.now())))
					f.write(u'## 测试题目`%s`，编译失败代码名称`%s.%s`\n' % (prob['name'], name, lang))
					try:
						f.write(u'编译器输出的stdout为：\n')
						f.write(open('tmp/stdout').read())
					except Exception as e:
						f.write(u'编译器没有输出stdout，具体错误为：\n')
						f.write(str(e) + '\n')
					try:
						f.write(u'编译器输出的stderr为：\n')
						f.write(open('tmp/stderr').read())
					except Exception as e:
						f.write(u'编译器没有输出stderr，具体错误为：\n')
						f.write(str(e) + '\n')
				raise Exception('`' + name + '.' + lang + '` compile failed.')
			return lang
	else:
		raise Exception('Can\'t find source file.')

def test(prob, name):
	scores = []
	times = []
	reports = []
	if prob['type'] == 'program':
		try:
			lang = compile(prob, name)
		except Exception as e:
			for i in range(len(prob.test_cases) + len(prob.sample_cases)):
				scores.append(0.0)
				times.append(0.0)
				reports.append(str(e))
			return scores, times, reports
	all_cases = prob.test_cases + prob.sample_cases + prob.pre_cases
	for case in all_cases:
		print('Case %s:%s  ' % (case['key'], case), end = '\r')
		sys.stdout.flush()
		shutil.copy(case.full() + '.in', pjoin('tmp', 'in'))
		shutil.copy(case.full() + '.ans', pjoin('tmp', 'ans'))
		for fname in ('in', 'ans'):
			if base.system == 'Windows':
				base.unix2dos(pjoin('tmp', fname))
			else:
				base.dos2unix(pjoin('tmp', fname))
		if prob['type'] == 'program':
			os.chdir('tmp')
			ret, time = run(name, prob['time limit'], prob['memory limit'], 'in', 'out', base.runners[lang])
			os.chdir('..')
		elif prob['type'] == 'output':
			if os.path.exists(pjoin('tmp', case['case'] + '.out')):
				shutil.copy(pjoin('tmp', case['case'] + '.out'), pjoin('tmp', 'out'))
				ret = None
				time = 0.0
			else:
				ret = 'Output file does not exist.'
				time = 0.0
		else:
			log.error(u'错误的题目类型`%s`。' % prob['type'])
			raise Exception('problem type error.')
		if not ret:
			if not os.path.exists(pjoin('tmp', 'out')):
				ret = 'Output file does not exist.'
				time = 0.0
				score = 0.0
			elif prob.chk:
				shutil.copy(pjoin('bin', prob.route), pjoin('tmp', 'chk' + base.elf_suffix))
				open('100.0', 'w').write('100.0\n')
				os.system('%s %s %s %s 100.0 tmp/score tmp/info' % (
					pjoin('tmp', 'chk' + base.elf_suffix),
					pjoin('tmp', 'in'),
					pjoin('tmp', 'out'),
					pjoin('tmp', 'ans')
				))
				os.remove('100.0')
				try:
					arbiter_out = ('/' if base.system != 'Windows' else '') + 'tmp/_eval.score'
					f = open(arbiter_out)
					report = f.readline().strip()
					score = float(f.readline()) * 0.1
					f.close()
					shutil.remove(arbiter_out)
				except Exception as e:
					try:
						report = open('tmp/info').read().strip()
					except Exception as e:
						report = ''
					score = float(open('tmp/score').readline()) * .01
			else:
				ret = os.system('%s %s %s > log' % (
					base.diff_tool,
					pjoin('tmp', 'ans'),
					pjoin('tmp', 'out')
				))
				if ret == 0:
					score = 1.0
					report = 'ok'
					if prob['type'] != 'output' and time > prob['time limit']:
						score = 0.0
						report += '(but time out)'
				else:
					score = 0.0
					report = 'wa'
					if prob['type'] != 'output' and time > prob['time limit']:
						report += '(and time out)'
		else:
			score = 0.0
			report = ret
		while os.path.exists(pjoin('tmp', prob['name'] + '.out')):
			try:
				os.remove(pjoin('tmp', prob['name'] + '.out'))
			except:
				pass
		scores.append(score)
		times.append(time)
		reports.append(report)
	return scores, times, reports

def packed_score(scores, times, reports, score_map, prob):
	pscore = []
	ptime = []
	preport = []
	for datum in prob.data:
		pscore.append(datum.score * min((scores[score_map[i]] for i in datum['cases'])))
		ptime.append(sum((times[score_map[i]] for i in datum['cases'] if scores[score_map[i]] > 0)))
		preport.append('Total %.1f' % datum.score)
	pscore.append(sum(pscore))
	ptime.append(sum(ptime))
	preport.append('')
	return (pscore, ptime, preport)
	
def test_problem(prob):
	log.info(u'尝试评测题目`%s`。' % prob.route)
	if 'users' not in prob:
		log.warning(u'题目`%s`缺少`users`字段，没有找到出题人的程序。' % prob.route)
		log.info(u'你需要在工程中放置你的代码，然后使用`python -m tuack.gen code`搜索源代码，或直接配置`users`字段。')
		return
	if 'data' not in prob or len(prob.test_cases) == 0:
		log.warning(u'题目`%s`缺少`data`字段，找不到测试数据。' % prob.route)
		log.info(u'你需要在文件夹`%s`下按*.in和*.ans格式存放测试数据，' % pjoin(prob.path, 'data'))
		log.info(u'然后使用`python -m tuack.gen data`添加它们或直接配置`data`字段。')
	if 'samples' not in prob or len(prob.sample_cases) == 0:
		log.warning(u'题目`%s`缺少`samples`字段，找不到样例数据。' % prob.route)
		log.info(u'你需要在文件夹`%s`下按*.in和*.ans格式存放样例数据，' % pjoin(prob.path, 'down'))
		log.info(u'然后使用`python -m tuack.gen samples`添加它们或直接配置`samples`字段。')
	if 'pre' in prob and len(prob.pre_cases) == 0:
		log.warning(u'题目`%s`的`pre`字段为空，找不到预测试数据。' % prob.route)
		log.info(u'如果这道题目有预测试数据，你需要在文件夹`%s`下按*.in和*.ans格式存放预测试数据，' % pjoin(prob.path, 'pre'))
		log.info(u'然后使用`python -m tuack.gen pre`在文件夹`%s`下搜索预测试数据。' % (pjoin(prob.path, 'pre')))
		log.info(u'如果这道题目没有预测试数据，你需要删除`conf.*`中的`pre`字段。')
	if len(prob.data) == 1 and prob.data[0].get('score') != 100:
		log.warning(u'题目`%s`的`data`字段只有一个元素，可能没有正确配置数据。' % prob.route)
		log.info(u'如果是一个包的CPC赛制，需要在这个元素中配置`score`项为100。')
		log.info(u'如果是多点或多子任务的赛制，需要将每类数据分类配置或打包配置。')
		log.info(u'详情请见%s。' % base.wiki(u'配置题目#数据'))
	log.info(u'共%d组样例，%d个预测试点，%d个测试点，%s打包评测%s。' % (
		len(prob.sample_cases),
		len(prob.pre_cases),
		len(prob.test_cases),
		u'是' if prob.packed else u'不是',
		(u'（共%d个包）' % len(prob.data) if len(prob.data) != 1 else u'（看样子是一个包的CPC赛制）') if prob.packed else ''
	))

	prob_failed = False

	case_features = [[] for i in range(len(prob.data))]

	with open(pjoin('result', prob.route) + '.csv', 'w') as fres:
		fres.write('%s,%s%s,summary,sample%s,pre%s\n' % (
			prob['name'],
			','.join(prob.test_cases),
			',' + ','.join(map(lambda datum : '{' + ';'.join(map(str, datum['cases'])) + '}', prob.data)) \
				if prob.packed else '',
			','.join(prob.sample_cases),
			','.join(prob.pre_cases)
		))
		for user, algos in prob.users().items():
			if not prob.all:
				match = base.any_prefix(rjoin(prob.route, user))
				if not match:
					continue
			for algo, algo_obj in algos.items():
				path = algo_obj['path']
				expe = algo_obj.get('expected')
				if not expe or len(expe) == 0:
					log.warning(u'程序`%s/%s`没有设置期望得分。' % (user, algo))
					log.info(u'如果该程序是标程、部分分程序或骗分程序等，需要设置它的`expected`字段。')
					log.info(u'有多种设置的格式，例如：`== 60`表示该程序应该得到60分。')
					log.info(u'如果该程序是其他功能程序，例如数据生成器，请将该程序从配置文件中移除。')
				if (not prob.all and match != 1 and not base.any_prefix(rjoin(prob.route, user, algo))):
					continue
				while os.path.exists('tmp'):
					try:
						shutil.rmtree('tmp')
					except:
						time.sleep(1e-2)
				if prob['type'] == 'program':
					while True:
						try:
							os.makedirs('tmp')
						except:
							time.sleep(1e-2)
						else:
							break
					try:
						shutil.copy(path, pjoin('tmp', path.split('/')[-1]))
					except Exception as e:
						log.error(u'程序`%s`没有找到。' % path)
						log.info(e)
						log.info(u'如果该程序是测试程序，请确保该程序存在；如果不是，请将其从`conf.*`中移除。')
						continue
				elif prob['type'] == 'output':
					for i in range(20):
						try:
							shutil.copytree(path, 'tmp')
						except Exception as e:
							time.sleep(1e-2)
							log.error(u'测试输出文件包`%s`没有找到或复制失败，正在重试 %d / 20。' % (path, i + 1))
							log.info(e)
						else:
							break
					else:
						log.error(u'跳过该输出文件包。')
						log.info(u'如果该文件包是测试输出，请确保该文件包存在；如果不是，请将其从`conf.*`中移除。')
				log.info(u'测试程序 %s:%s:%s' % (prob['name'], user, algo))
				scores, times, reports = test(prob, path.split('/')[-1].split('.')[0])
				while os.path.exists('tmp'):
					try:
						shutil.rmtree('tmp')
					except:
						pass
				tc = len(prob.test_cases)
				score_map = {}
				for i in range(tc):
					score_map[prob.test_cases[i]] = i
				if prob.packed:
					packed = packed_score(scores[:tc], times[:tc], reports[:tc], score_map, prob)
					tot = sum(packed[0][:-1])
					scores = scores[:tc] + packed[0] + scores[tc:]
					times = times[:tc] + packed[1] + times[tc:]
					reports = reports[:tc] + packed[2] + reports[tc:]
					for i, s in enumerate(packed[0][:-1]):
						case_features[i].append(s)
				elif tc > 0:
					ratio = 100. / tc
					scores = [score * ratio for score in scores[:tc] + [sum(scores[:tc])] + scores[tc:]]
					packed = packed_score(scores[:tc], times[:tc], reports[:tc], score_map, prob)
					tot = sum(scores[:tc])
					times = times[:tc] + [sum((val for idx, val in enumerate(times[:tc]) if scores[idx] > 0))] + times[tc:]
					reports = reports[:tc] + [''] + reports[tc:]
					for i, s in enumerate(packed[0][:-1]):
						case_features[i].append(s)
				else:
					# FIXME: Do I do this?
					tot = sum(scores)
					scores = [0.0, 0.0] + scores
					times = [0.0, 0.0] + times
					reports = ['', ''] + reports

				if not prob.expect(user, algo, tot):
					log.error(u'未达到预期。预期分数 %s ，实际得分 %.2f' % (algo_obj['expected'], tot))
					prob_failed = True

				scores = map(lambda i : '%.2f' % i, scores)
				times = map(lambda i : '%.3f' % i, times)
				reports = map(lambda i : i.replace('\n', '\\n').replace(',', ';').replace('\r', ''), reports)
				for title, line in [(user, scores), (algo, times), ('', reports)]:
					fres.write('%s,%s\n' % (title, ','.join(line)))
	used_features = {}
	for i, f in enumerate(case_features):
		tf = tuple(f)
		if tf in used_features:
			log.warning(u'没有程序区分数据包%s和%s。' % (prob.data[used_features[tf]]['cases'], prob.data[i]['cases']))
			log.info(u'对每种不同类型的数据，都应当有程序可以被这些数据区分。')
		else:
			used_features[tf] = i
	if base.start_file:
		base.xopen_file(pjoin('result', prob.route) + '.csv')
	return not prob_failed

def test_progs():
	test_failed = False
	if base.conf.folder != 'problem' and not os.path.exists('result'):
		os.makedirs('result')
	for day in base.days():
		path = pjoin('result', day.route)
		if not os.path.exists(path):
			os.makedirs(path)
	for prob in base.probs():
		try:
			if not test_problem(prob):
				test_failed = True
		except Exception as e:
			log.error(traceback.format_exc())
	return not test_failed

if __name__ == '__main__':
	try:
		if base.init():
			base.work = 'test'
			if base.do_pack:
				from . import packer
				base.run_exc(packer.test)
			ex, success = base.run_exc(test_progs)
			if ex or not success:
				sys.exit(1)
		else:
			log.info(u'这是测试出题人数据和程序的测试器，测试器没有细分的工作。')
			sys.exit(1)
	except base.NoFileException as e:
		log.error(e)
		log.info(u'尝试使用`python -m tuack.gen -h`获取如何生成一个工程。')
		sys.exit(1)



# -*- coding: utf-8 -*-

from __future__ import print_function
import os
import re
import sys
import json
import datetime
import shutil
import subprocess
import time
import signal
import zipfile
from multiprocessing import Process, Queue
from functools import wraps
from threading import Timer
import platform
from . import base
from .base import log

def tsinsen_oj():
	new_tc = []
	
	class Base(object):
		def __init__(self):
			pass
		def write(self, text):
			pass
		def close(self):
			pass
	class Conf(Base):
		def __init__(self, *names):
			self.names = names
			self.buff = b''
		def write(self, text):
			self.buff += text
		def close(self):
			if type(self.buff) == bytes:
				cur = self.buff.decode('utf-8').strip()
			else:
				cur = self.buff
			for name in reversed(self.names[1:]):
				if not name:
					cur = [cur]
				else:
					cur = {name : cur}
			base.conf[self.names[0]] = cur
	class File(Base):
		def __init__(self, *names):
			path = ''
			for name in names[:-1]:
				path = base.pjoin(path, name)
				if not os.path.exists(path):
					os.makedirs(path)
			self.f = open(base.pjoin(*names), 'wb')
		def write(self, text):
			self.f.write(text)
		def close(self):
			self.f.close()
	class HtmlFile(Base):
		def __init__(self, *names):
			path = ''
			for name in names[:-1]:
				path = base.pjoin(path, name)
				if not os.path.exists(path):
					os.makedirs(path)
			self.fname = base.pjoin(*names)
			self.buff = b''
		def write(self, text):
			self.buff += text.strip() + b'\n'
		def close(self):
			ret = json.dumps(json.dumps(self.buff.decode('utf-8'))).encode('utf-8')
			f = open(self.fname, 'wb')
			f.write(b'{{ self.title() }}\n')
			f.write(b'{{ render(%s, \'html\') }}\n' % ret)
			f.close()
	class FileName(Base):
		last_name = None
		def write(self, text):
			FileName.last_name = text.decode('utf-8').rstrip('\n\r')
	class Base64File(Base):
		def __init__(self):
			self.buff = b''
		def write(self, text):
			self.buff += text
		def close(self):
			import base64
			if not os.path.exists('resources'):
				os.makedirs('resources')
			open(base.pjoin('resources', FileName.last_name), 'wb').write(base64.b64decode(self.buff))
	Title = lambda : Conf('title', 'zh-cn')
	CheckPoint = lambda : Conf('key words')
	TestMethod = Base
	InputFileName = Base
	OutputFileName = Base
	Description = lambda : HtmlFile('statement', 'zh-cn.md')
	Judger = lambda : File('data', 'chk', 'chk.cpp')
	
	class TimeLimit(Conf):
		def __init__(self):
			super(TimeLimit, self).__init__('time limit')
		def close(self):
			self.buff = float(self.buff.strip()[:-1])
			super(TimeLimit, self).close()
	
	class MemoryLimit(Conf):
		def __init__(self):
			super(MemoryLimit, self).__init__('memory limit')
		def close(self):
			if len(self.buff) >= 2:
				s = self.buff.strip().decode('utf-8')
				sp = 1 if '0' <= s[-2] <= '9' else 2
				self.buff = (s[:-sp] + ' ' + s[-sp:]).encode('utf-8')
			super(MemoryLimit, self).close()
	
	class Solution(File):
		def __init__(self):
			super(Solution, self).__init__('tsinsen-oj', 'std', 'std.cpp')
		def close(self):
			super(Solution, self).close()
			if 'users' not in base.conf:
				base.conf['users'] = {}
			if 'tsinsen-oj' not in base.conf['users']:
				base.conf['users']['tsinsen-oj'] = {}
			base.conf['users']['tsinsen-oj']['std'] = 'tsinsen-oj/std/std.cpp'
			
	class InData(File):
		last_in = None
		def __init__(self):
			tc = set(base.conf.test_cases)
			for i in range(1, len(tc) + 2):
				if str(i) not in tc:
					name = i
			new_tc.append(name)
			base.conf.test_cases.append(name)
			InData.last_in = name
			super(InData, self).__init__('data', '%d.in' % name)
			
	class OutData(File):
		def __init__(self):
			super(OutData, self).__init__('data', '%d.ans' % InData.last_in)
			
	class Checkers(File):
		fname = ('data', 'chk', 'chk.cpp')
		def __init__(self):
			super(Checkers, self).__init__(*Checkers.fname)
		def close(self):
			super(Checkers, self).close()
			if os.path.getsize(base.pjoin(*Checkers.fname)) < 10:
				shutil.rmtree(base.pjoin(*Checkers.fname[:2]), ignore_errors = True)
	import re
	key_re = re.compile('(\w*)\((\w*)\)')

	if base.conf.folder != 'problem':
		log.error(u'只能导入到一个problem的工程中。')
		return
	status = None
	for line in open(base.args[0], 'rb'):
		if not status:			# 啥都没有
			try:
				key = line.decode('utf-8').strip()[:-1]
				m = key_re.match(key)
				if m:
					if m.group(1) == 'File':
						log.warning(u'题面含有文件，请在题面中手工查找字符串`%s`并进行相应修改，该文件保存在`resources/%s`。' % (
							m.group(2), FileName.last_name
						))
						buffer = Base64File()
					else:
						log.warning(u'无法解析关键词`%s`。' % key)
				else:
					buffer = eval(key)()
			except Exception as e:
				log.warning(u'无法解析关键词`%s`。' % key)
				log.debug(e)
				buffer = Base()
			status = True
		elif status == True:	# 读到了等号的行，还没读到标记开始的行
			status = line
		elif status == line:	# 读到了标记结束的行
			buffer.close()
			status = None
		else:					# 读到了标记开始的行，还没读到标记结束的行
			buffer.write(line)
	if len(new_tc) > 0:
		if 'data' not in base.conf:
			base.conf['data'] = []
		base.conf['data'].append({'cases' : new_tc})
	base.save_json(base.conf)

work_list = {
	'tsinsen-oj' : tsinsen_oj
}

if __name__ == '__main__':
	try:
		if base.init() and len(base.args) != 0:
			for base.work in base.works:
				base.run_exc(work_list[base.work])
		else:
			log.info(u'这个工具用于导入其他类型的工程，参数1必须是来源路径。')
			log.info(u'支持的工作：%s' % ','.join(sorted(work_list.keys())))
	except base.NoFileException as e:
		log.error(e)
		log.info(u'尝试使用`python -m tuack.gen -h`获取如何生成一个工程。')



# -*- coding: utf-8 -*-

from __future__ import print_function
import os
import re
import sys
import json
import datetime
import shutil
import subprocess
import time
import signal
import zipfile
from multiprocessing import Process, Queue
from functools import wraps
from threading import Timer
import platform
from . import base
from .base import log, pjoin

vars_re = r"\{%-? *do *vars\.__setitem__\( *'sample_id' *, *(\d+) *\) -%\}"

title_choices = {
	'sample' : {
		u'样例(\\d+)?',
		u'输入输出样例(\\d+)?',
		u'样例输入输出(\\d+)?',
		u'测试用例(\\d+)?',
		r'sample\s*(\d+)?',
		vars_re
	},
	'sample input' : {
		u'样例(\\d+)?输入(\\d+)?',
		u'输入(\\d+)?样例(\\d+)?',
		r'sample\s*input\s*(\d+)?',
		r'input\s*sample\s*(\d+)?'
	},
	'sample output' : {
		u'样例(\\d+)?输出(\\d+)?',
		u'输出(\\d+)?样例(\\d+)?',
		r'sample\s*output\s*(\d+)?',
		r'output\s*sample\s*(\d+)?'
	},
	'sample explanation' : {
		u'样例(\\d+)?说明(\\d+)?',
		u'样例(\\d+)?解释(\\d+)?',
		r'sample\s*explanation\s*(\d+)?'
	},
	'input' : {
		u'输入',
		'input'
	},
	'outout' : {
		u'输出',
		'output'
	},
	'explanation' : {
		u'解释',
		u'说明',
		'explanation'
	},
	'input format' : {
		u'输入格式',
		u'输入说明',
		u'输入要求',
		'input format'
	},
	'output format' : {
		u'输出格式',
		u'输出说明',
		u'输出要求',
		'output format'
	},
	'description' : {
		u'题目描述',
		u'题目说明',
		u'题目',
		u'问题描述',
		u'问题说明',
		u'问题',
		'description',
		'statement',
	},
	'background' : {
		u'题目背景',
		u'背景',
		'background',
	},
	'hint' : {
		u'提示',
		u'温馨提示',
		u'hint'
	},
	'subtasks' : {
		u'(数据规模|要求|限制|约束|约定|子任务)(与(数据规模|要求|限制|约束|约定|子任务))?',
		r'subtask(s?)'
	},
	'scoring' : {
		u'评分方式',
		'scoring'
	}
}

new_re_symbols = {
	r'\C' : '[\u4e00-\u9fbf]',
	r'\P' : '[\uff00-\uffef\u3000-\u303f\u2000-\u203f]'
}

text_patterns = {
	u'中文和英文、字符串或公式中间没有空格' : r'\C(?:\w|`|\$)',
	u'英文、字符串或公式和中文中间没有空格' : r'(?:\w|`|\$)\C',
	u'中文之间有空格' : r'\C \C',
	u'使用了英文标点或不在公式中的运算符号' : '[,\\.\\?\\:;\'"\\[\\]\\{\\}!\\\\\\|\\(\\)\\+=]'
}

title_re = re.compile('^(' + '|'.join(['|'.join(value) for value in title_choices.values()]) + ')$')
std_title_re = re.compile('^\{\{ *s\( *\'(.*)\' *(( *, *[-+.\'\"a-zA-Z0-9_ ]+)*)\) *\}\} *$')
doc_title_re = u'^#* *( *【|\{\{ *_ *\( *\')?(%s)(\' *\) *\}\}|】)?( *\{#.*\})? *#* *$'
html_equation_re = re.compile(r'\\\[(.*)\\\]\{\.math \.inline\}')
format_log_re = re.compile(r'^(\w) (\d+) (.*)$')

math_trans = {
	'*' : '',
	u'≤' : r' \le',
	u'≥' : r' \ge',
	u'←' : r' \leftarrow',
	u'→' : r' \rightarrow',
	u'×' : r' \times',
	u'∑' : r' \sum',
	r'\$' : ''
}

def is_digit(s):
	for c in s:
		if not '0' <= c <= '9':
			return False
	return True

def is_english(s):
	for c in s:
		if not('a' <= c <= 'z' and 'A' <= c <= 'Z'):
			return False
	return True

def is_chinese(s):
	for c in s:
		if not u'\u4e00' <= c <= u'\u9fbf':
			return False
	return True

def get_text(s):
	ret = ''
	for c in s:
		if is_chinese(c) or is_english(c) or is_digit(c):
			ret += c
	return ret.lower()
	
def html_math2tex_math(s):
	ret = ''
	i = 0
	l = len(s)
	in_sup = False
	in_sub = False
	while True:
		if i == l:
			break
		if i + 1 < l:
			flag = False
			if s[i:i+2] == '\*':
				ret += '*'
			elif s[i:i+2] == '\^':
				if in_sup:
					ret += ' }'
				else:
					ret += '^{ '
				in_sup ^= True
			elif s[i:i+2] == '\~':
				if in_sub:
					ret += ' }'
				else:
					ret += '_{ '
				in_sub ^= True
			elif s[i:i+2] in math_trans:
				ret += math_trans[s[i:i+2]]
			else:
				flag = True
			if not flag:
				i += 2
				continue
		if s[i] in math_trans:
			ret += math_trans[s[i]]
		else:
			ret += s[i]
		i += 1
	return ret

class Section:
	used = {'input' : 0, 'output' : 0}
	def __init__(self, name = None, args = None):
		self.name = name
		if args == None:
			self.args = []
		else:
			self.args = args
		self.lines = []
	def clear_paragraph(self):
		flag = True
		for idx, line in enumerate(self.lines):
			if idx % 2 == 1 and line.rstrip('\n\r') != '':
				flag = False
		if flag:
			self.lines = [line for idx, line in enumerate(self.lines) if idx % 2 == 0]
	def write(self, f, idx = None):
		if self.name == 'sample':
			if len(self.lines) == 0:
				return
			else:
				self.args.append('')
		for name, suf in [('input', '.in'), ('output', '.ans')]:
			if self.name == 'sample ' + name or self.name == name:
				self.used[name] += 1
				self.clear_paragraph()
				with open(pjoin(base.prob.path, 'down', str(self.used[name]) + suf), 'w') as ff:
					if len(self.lines) > 1 and self.lines[0].startswith('```'):
						lines = self.lines[1:-1]
					else:
						lines = self.lines
					for line in lines:
						ff.write(line + '\n')
				if name == 'input':
					f.write(b'{{ s(\'sample\', %d) }}\n\n{{ self.sample_text() }}\n\n' % self.used[name])
				return
		if idx == 0:
			f.write(b'{{ self.title() }}\n')
		if idx == 0 and self.name and self.name not in title_choices:
			base.prob['title'][base.prob.lang()] = self.name.strip()
		elif self.name:
			f.write(("\n{{ s('%s'%s) }}\n\n" % (
				self.name,
				', ' + ', '.join(map(json.dumps, self.args)) if len(self.args) > 0 else ''
			)).encode('utf-8'))
		for line in self.lines:
			f.write((line + '\n').encode('utf-8'))
	def is_empty(self):
		return not self.name and len(self.lines) == 0
	def format_line(self, line):
		ret = line
		while True:
			m = html_equation_re.search(ret)
			if not m:
				break
			while True:
				sm = m
				l, r = sm.span()
				sub = sm.string[l:r - 1]
				m = html_equation_re.search(sub)
				if m:
					continue
				sub = sm.string[l:r]
				res = sm.group(1)
				ret = ret.replace(sub, '$%s$' % html_math2tex_math(res))
				break
		return ret
	def format(self):
		self.lines = [self.format_line(line) for line in self.lines]

def sure_title(line):
	inp = line
	if inp.startswith('##'):
		return True
	if inp.startswith(u'【') and inp.endswith(u'】'):
		return True
	if std_title_re.match(inp):
		return True
	if re.match(vars_re, inp):
		return True
	txt = get_text(inp)
	if title_re.match(txt):
		return True
	return False

def get_title(line):
	m = std_title_re.match(line)
	if m:
		gps = m.groups()
		name = gps[0]
		args = json.loads('[' + ','.join(gps[1].split(',')[1:]) + ']')
		return name, args
	m = re.match(doc_title_re % '[-+.\'\"a-zA-Z0-9_ \u4e00-\u9fa5\u3040-\u309f\u30a0-\u30ff]+', line.lower())
	if m:
		title = m.group(2)
		m = title_re.match(title)
		if m:
			for key, pts in title_choices.items():
				for pt in pts:
					m = re.match('^' + pt + '$', title)
					if m:
						name = key
						args = []
						if key.startswith('sample'):
							for s in m.groups():
								if s and s != '':
									args = [json.loads(s)]
									log.debug(args)
									break
						return name, args
		return title, []
	return line, []

def to_sections(lines):
	def clear(buff):
		if buff != []:
			cur.lines.append(' '.join(buff))
			cur.lines.append('')
			buff = []
		return buff
	sections = []
	cur = Section()
	buff = []
	in_quote = False
	in_quote_space = False
	in_quote_tab = False
	in_equation = False
	in_table = False
	last_title = 0
	for line in lines:
		if in_quote_space:
			if line.startswith('    '):
				cur.lines.append(line.lstrip(' '))
				continue
			else:
				in_quote_space = False
				cur.lines.append('```')
		elif in_quote_tab:
			if line.startswith('\t'):
				cur.lines.append(line.lstrip(' '))
				continue
			else:
				in_quote_tab = False
				cur.lines.append('```')
		elif in_table:
			if line.startswith('  '):
				cur.lines.append(line)
				continue
			else:
				in_table = False
		if in_quote:
			if line == '```':
				in_quote = False
			cur.lines.append(line)
		elif in_equation:
			if line == '$$':
				in_quote = False
			cur.lines.append(line)
		else:
			last_title = last_title - 1 if last_title >= 1 else 0
			if line.startswith('```'):
				in_quote = True
				buff = clear(buff)
				cur.lines.append(line)
			elif line.strip() == '$$':
				in_equation = True
				buff = clear(buff)
				cur.lines.append(line)
			elif sure_title(line):
				buff = clear(buff)
				title, args = get_title(line)
				if not cur.is_empty():
					sections.append(cur)
				cur = Section(title, args)
				last_title = 2
			elif line.startswith('    '):
				in_quote_space = True
				buff = clear(buff)
				cur.lines.append('```')
				cur.lines.append(line.lstrip(' '))
			elif line.startswith('\t'):
				in_quote_space = True
				buff = clear(buff)
				cur.lines.append('```')
				cur.lines.append(line.lstrip(' '))
			elif line.startswith('  '):
				in_table = True
				buff = clear(buff)
				cur.lines.append(line)
			elif line.startswith('------'):
				if last_title >= 1:
					continue
				if buff != []:
					buff.pop()
				buff = clear(buff)
				title, args = get_title(last)
				if not cur.is_empty():
					sections.append(cur)
				cur = Section(title, args)
			elif line == '':
				buff = clear(buff)
			else:
				buff.append(line)
		last = line
	buff = clear(buff)
	if not cur.is_empty():
		sections.append(cur)
	return sections

def format():
	for base.prob in base.probs():
		prob = base.prob
		path = pjoin(base.conf.path, 'statement', base.conf.lang() + '.md')
		lines = [line.rstrip(b'\r\n').decode('utf-8') for line in open(path, 'rb').readlines()]
		sections = to_sections(lines)
		for section in sections:
			section.format()
		with open(path, 'wb') as f:
			for idx, section in enumerate(sections):
				section.write(f, idx)
	base.save_json(base.conf)

def format_check_one(path):
	log.info(u'检查文件`%s`。' % path)
	widths = [
		(126, 1), (159, 0), (687, 1), (710, 0), (711, 1), 
		(727, 0), (733, 1), (879, 0), (1154, 1), (1161, 0), 
		(4347, 1), (4447, 2), (7467, 1), (7521, 0), (8369, 1), 
		(8426, 0), (9000, 1), (9002, 2), (11021, 1), (12350, 2), 
		(12351, 1), (12438, 2), (12442, 0), (19893, 2), (19967, 1),
		(55203, 2), (63743, 1), (64106, 2), (65039, 1), (65059, 0),
		(65131, 2), (65279, 1), (65376, 2), (65500, 1), (65510, 2),
		(120831, 1), (262141, 2), (1114109, 1),
	]
	def get_width(o):
		if o == '\t':
			return 4
		o = ord(o)
		if o == 0xe or o == 0xf:
			return 0
		for num, wid in widths:
			if o <= num:
				return wid
		return 1
	def output():
		level = {
			'E' : log.error,
			'W' : log.warning,
			'I' : log.info,
			'D' : log.debug,
		}[logs[idx][0]]
		level(u'第%d行，第%d个字：%s' % (lineno, col, logs[idx][2]))
		lef = col - 10 if col > 10 else 0
		rig = min(lef + 21, len(code.rstrip()))
		log.info(code[lef:rig].replace('\u200b', '?').replace('\t', '    '))
		lef_cur = 0
		for c in code[:lef]:
			lef_cur += get_width(c)
		log.info(' ' * (cur - lef_cur) + u'↑')
	os.system('%s < %s > format.tmp 2> format.log' % (
		pjoin(base.tool_path, base.format_checker_name),
		path
	))
	tot = 0
	logs = []
	for line in open('format.log', 'rb'):
		m = format_log_re.match(line.decode('utf-8').rstrip())
		if m:
			logs.append((m.group(1), int(m.group(2)), m.group(3)))
	if len(logs) == 0:
		return
	idx = 0
	for lineno, line in enumerate(open(path, 'rb'), 1):
		code = line.decode('utf-8')
		cur = 0
		for col, ch in enumerate(code):
			tot += len(ch.encode('utf-8'))
			if tot >= logs[idx][1]:
				output()
				idx += 1
			if idx >= len(logs):
				return
			cur += get_width(ch)

def format_check():
	base.check_install('format')
	for base.prob in base.probs():
		prob = base.prob
		path = pjoin(prob.path, 'statement', 'zh-cn.md')
		format_check_one(path)

def load():
	if len(base.args) != 1:
		log.error(u'无法转换，必须传入恰好一个参数。')
		sys.exit(1)
	if base.conf.folder != 'problem':
		log.error(u'只能处理单个题目，请到题目目录下进行操作。')
		sys.exit(1)
	os.system('pandoc %s -o %s' % (base.args[0], pjoin(base.conf.path, 'statement', base.conf.lang() + '.md')))

class_list = {
	'load' : load,
	'format' : format,
	'check' : format_check
}

if __name__ == '__main__':
	try:
		if base.init():
			base.check_install('pandoc')
			base.check_install('jinja2')
			for base.work in base.works:
				for base.lang in base.langs:
					base.run_exc(class_list[base.work])
		else:
			log.info(u'这是用来处理题面的工具。')
			log.info(u'支持的工作：')
			log.info(u'  load     必须包含一个参数输入文件名，表示要导入的题面。')
			log.info(u'  format   尝试对当前的中文题面进行简单的格式化，危险操作请注意备份。')
			log.info(u'  check    对题面进行格式检查。')
			sys.exit(1)
	except base.NoFileException as e:
		log.error(e)
		log.info(u'尝试使用`python -m tuack.gen -h`获取如何生成一个工程。')
		sys.exit(1)




from setuptools import setup, find_packages
from setuptools.command.develop import develop
from setuptools.command.install import install
from subprocess import check_call
import platform

def check():
	system = platform.system()
	if system == 'Linux':
		check_call("sudo apt-get install pandoc git git-lfs".split())
		check_call("git lfs install".split())
	elif system == 'Darwin':
		check_call("brew update".split())
		check_call("brew install pandoc".split())
		check_call("brew install git".split())
		check_call("brew install git-lfs".split())
		check_call("git lfs install".split())

class PostDevelopCommand(develop):
	"""Post-installation for development mode."""
	def run(self):
		#check()
		develop.run(self)

class PostInstallCommand(install):
	"""Post-installation for installation mode."""
	def run(self):
		#check()
		install.run(self)

requires = [
	'jinja2 >= 2.8',
	'natsort >= 5.0.3',
	'pyyaml >= 3.11'
]

setup(
	name = 'tuack',
	version = '0.1.4.5',
	packages = find_packages(),
	author = 'Chen Xumin, Zhang Ruizhe, Liu Xiaoyi, Chen Junkun',
	author_email = 'chen.xm.mu@gmail.com, 657228726@qq.com, circuitcoder0@gmail.com, 1261954105@qq.com',
	url = '',
	license = 'https://git.thusaac.org/publish/tuack/blob/master/LICENSE',
	description = 'Tools for generating an Tsinghua/OI/ICPC-styled problem or contest for multiple judges.',
	cmdclass={
		'develop': PostDevelopCommand,
		'install': PostInstallCommand,
	},
	#requires = requires,
	install_requires = requires,
	setup_requires = requires,
	package_data = {
		'tuack': ['templates/*.*', 'templates/*/*/*', 'sample/*.*', 'sample/*/*', 'lex/*.*', 'lex/*']
	}
)




# -*- coding: utf-8 -*-

from __future__ import print_function
import os
import re
import sys
import json
import datetime
import shutil
import subprocess
import time
import signal
import zipfile
from multiprocessing import Process, Queue
from functools import wraps
from threading import Timer
import platform
from . import base
from .base import log, pjoin, rjoin
import traceback
import __main__

def format():
	fname = base.format_checker_name
	if base.system != 'Windows':
		install_path = '/'.join(__main__.__file__.split('/')[:-1]) + '/lex'
	else:
		install_path = __main__.__file__ + '\\..\\lex'
	shutil.copy(pjoin(install_path, fname), base.tool_path)
	fpath = pjoin(base.tool_path, fname)
	if base.system != 'Windows':
		os.system('chmod +x %s' % fpath)
		ret = os.system('%s -v' % fpath)
	else:
		ret = os.system('%s -v' % fpath)
	if ret == 0:
		log.info(u'`format`安装成功。')
		return
	log.info(u'直接安装format checker失败，请尝试用以下方式手工编译：')
	log.info(u'1. 先安装flex和bison两个工具。')
	log.info(u'2. 在目录tuack/lex下运行`python compile.pyinc`或在其他地方运行后将编译好的文件复制进该目录。')
	log.info(u'3. 重新运行`python -m tuack.install format`。')

def pre_check(name):
	def runner():
		log.info(u'尝试安装`%s`。' % name)
		'''
		try:
			base.check_install(name)
			log.info(u'`%s`已安装，跳过安装。' % name)
			return
		except Exception as e:
		'''
		eval(name)()
	return runner

packages = ['format']
work_list = {key : pre_check(key) for key in packages}
work_list['full'] = lambda : [work_list[key]() for key in work_list if key not in {'full', 'std'}]
work_list['std'] = lambda : [work_list[key]() for key in ('format',)]

if __name__ == '__main__':
	try:
		result = base.init()
	except base.NoFileException as e:
		base.conf = None
		result = True
	if result:
		for base.work in base.works:
			base.run_exc(work_list[base.work])
	else:
		log.info(u'这个工具用于引导安装tuack的组件。')
		log.info(u'支持的工作：')
		log.info(u'  std      标准安装，包含常用小型工具。')
		log.info(u'  full     完整安装，包含所有工具，可能会很大。')
		log.info(u'  format   题面格式检查器。')



# -*- coding: utf-8 -*-

from datetime import *
from math import *
import sys
if sys.version_info >= (3, 0): 
	from builtins import *
else:
	from __builtin__ import *
import json
from .num2chinese import *

def int_lg(num):
	if num == 0:
		return float('-inf')
	if num < 0:
		return float('nan')
	if num >= 10:
		if type(num) == int:
			return int_lg(num // 10) + 1
		else:
			return int_lg(num / 10) + 1
	if num < 1:
		return int_lg(num * 10) - 1
	return 0

def comma(num):
	if type(num) != int:
		return str(num)
	if num < 0:
		return '-' + comma(-num)
	if num < 1000:
		return str(num)
	return comma(num // 1000) + ',' + '%03d' % (num % 1000)
	
def hn(num, style = None):
	'''
	适合阅读的数
	'''
	ret = ''
	l = 0
	if num == 0:
		return '0'
	if num < 0:
		neg = '-'
		num = -num
	else:
		neg = ''
	n = int_lg(num)
	if num != 10 ** n:
		if num // 10 ** n * 10 ** n == num:
			ret += str(num // 10 ** n)
			l += 1
		else:
			tmp = str(float(num) / 10 ** n)
			ret += tmp
			l += len(tmp)
		ret += ' \\times '
		l += 1
	ret += '10^{%d}' % n
	l += 2 + len(str(n))
	if style == 'x':
		ret = ret
	elif style == ',':
		ret = comma(num)
	else:
		ret = (comma(num) if l >= len(str(num)) * 4 // 3 else ret)
	return neg + ret
	
def js_hn(num):
	'''
	适合阅读的数，json版本
	'''
	return json.dumps(hn(num))[1:-1]

def to_time(dttm):
	try:
		return datetime.strptime(dttm, '%Y-%m-%d %H:%M:%S')
		#return datetime.strptime(dttm, '%Y-%m-%d %H:%M:%S%z')	#python2 兼容性不好，暂时不用时区了
	except Exception as e:
		return datetime.strptime(dttm[:-5], '%Y-%m-%d %H:%M:%S')
	
def time_range(start, end, year = '-', month = '-', day = ''):
	st = datetime.strptime(start[:-5], '%Y-%m-%d %H:%M:%S')
	ed = datetime.strptime(end[:-5], '%Y-%m-%d %H:%M:%S')
	#st = datetime.strptime(start, '%Y-%m-%d %H:%M:%S%z')	#python2 兼容性不好，暂时不用时区了
	#ed = datetime.strptime(end, '%Y-%m-%d %H:%M:%S%z')
	ret = st.strftime('%Y%%sQAQ%m%%sQAQ%d%%s %H:%M').replace('QAQ0', '').replace('QAQ', '') % (year, month, day)
	if st.second:
		ret += ':%02d' % st.second
	ret += ' ~ '
	flag = False
	if flag or ed.year != st.year:
		flag = True
		ret += '%04d%s' % (ed.year, year)
	if flag or ed.month != st.month:
		flag = True
		ret += '%d%s' % (ed.month, month)
	if flag or ed.day != st.day:
		flag = True
		ret += '%d%s' % (ed.day, day)
	ret += ' %02d:%02d' % (ed.hour, ed.minute)
	if ed.second:
		ret += ':%02d' % ed.second
	return ret

def json_dumps_twice(s):
	return json.dumps(json.dumps(s))

class Args(object):
	def __init__(self, *data):
		self.data = data
	def get(self, key, d = None, skip_none = False):
		return (j['args'].get(key, d) for i in self.data for j in i if not skip_none or key in j['args'] != None)
	def keys(self):
		ky = set()
		for i in self.data:
			ky |= set(i.keys())
		return ky
	def __getitem__(self, key):
		return self.get(key)
	def sum(self, key, d = None):
		return sum(self.get(key, d, True))
	def min(self, key, d = None):
		return min(self.get(key, d, True))
	def max(self, key, d = None):
		return max(self.get(key, d, True))

a = Args
